package com.philkes.notallyx.data.model

import androidx.room.TypeConverter
import org.json.JSONArray
import org.json.JSONException
import org.json.JSONObject

object Converters {

    @TypeConverter fun labelsToJson(labels: List<String>) = JSONArray(labels).toString()

    @TypeConverter fun jsonToLabels(json: String) = JSONArray(json).iterable<String>().toList()

    @TypeConverter
    fun filesToJson(files: List<FileAttachment>): String {
        val objects =
            files.map { file ->
                val jsonObject = JSONObject()
                jsonObject.put("localName", file.localName)
                jsonObject.put("originalName", file.originalName)
                jsonObject.put("mimeType", file.mimeType)
            }
        return JSONArray(objects).toString()
    }

    @TypeConverter
    fun jsonToFiles(json: String): List<FileAttachment> {
        val iterable = JSONArray(json).iterable<JSONObject>()
        return iterable.map { jsonObject ->
            val localName = getSafeLocalName(jsonObject)
            val originalName = getSafeOriginalName(jsonObject)
            val mimeType = jsonObject.getString("mimeType")
            FileAttachment(localName, originalName, mimeType)
        }
    }

    @TypeConverter
    fun audiosToJson(audios: List<Audio>): String {
        val objects =
            audios.map { audio ->
                val jsonObject = JSONObject()
                jsonObject.put("name", audio.name)
                jsonObject.put("duration", audio.duration)
                jsonObject.put("timestamp", audio.timestamp)
            }
        return JSONArray(objects).toString()
    }

    @TypeConverter
    fun jsonToAudios(json: String): List<Audio> {
        val iterable = JSONArray(json).iterable<JSONObject>()
        return iterable.map { jsonObject ->
            val name = jsonObject.getString("name")
            val duration = jsonObject.getLong("duration")
            val timestamp = jsonObject.getLong("timestamp")
            Audio(name, duration, timestamp)
        }
    }

    @TypeConverter
    fun jsonToSpans(json: String): List<SpanRepresentation> {
        val iterable = JSONArray(json).iterable<JSONObject>()
        return iterable.map { jsonObject ->
            val bold = jsonObject.getSafeBoolean("bold")
            val link = jsonObject.getSafeBoolean("link")
            val italic = jsonObject.getSafeBoolean("italic")
            val monospace = jsonObject.getSafeBoolean("monospace")
            val strikethrough = jsonObject.getSafeBoolean("strikethrough")
            val start = jsonObject.getInt("start")
            val end = jsonObject.getInt("end")
            SpanRepresentation(bold, link, italic, monospace, strikethrough, start, end)
        }
    }

    @TypeConverter
    fun spansToJson(list: List<SpanRepresentation>) = spansToJSONArray(list).toString()

    @TypeConverter
    fun jsonToItems(json: String): List<ListItem> {
        val iterable = JSONArray(json).iterable<JSONObject>()
        return iterable.map { jsonObject ->
            val body = jsonObject.getString("body")
            val checked = jsonObject.getBoolean("checked")
            val isChild = jsonObject.getSafeBoolean("isChild")
            val order = jsonObject.getSafeInt("order")
            ListItem(body, checked, isChild, order, mutableListOf())
        }
    }

    @TypeConverter fun itemsToJson(list: List<ListItem>) = itemsToJSONArray(list).toString()

    fun itemsToJSONArray(list: List<ListItem>): JSONArray {
        val objects =
            list.map { item ->
                val jsonObject = JSONObject()
                jsonObject.put("body", item.body)
                jsonObject.put("checked", item.checked)
                jsonObject.put("isChild", item.isChild)
                jsonObject.put("order", item.order)
            }
        return JSONArray(objects)
    }

    fun spansToJSONArray(list: List<SpanRepresentation>): JSONArray {
        val objects =
            list.map { representation ->
                val jsonObject = JSONObject()
                jsonObject.put("bold", representation.bold)
                jsonObject.put("link", representation.link)
                jsonObject.put("italic", representation.italic)
                jsonObject.put("monospace", representation.monospace)
                jsonObject.put("strikethrough", representation.strikethrough)
                jsonObject.put("start", representation.start)
                jsonObject.put("end", representation.end)
            }
        return JSONArray(objects)
    }

    private fun getSafeLocalName(jsonObject: JSONObject): String {
        return try {
            jsonObject.getString("localName")
        } catch (e: JSONException) {
            jsonObject.getString("name")
        }
    }

    private fun getSafeOriginalName(jsonObject: JSONObject): String {
        return try {
            jsonObject.getString("originalName")
        } catch (e: JSONException) {
            getSafeLocalName(jsonObject).substringAfterLast("/")
        }
    }

    private fun JSONObject.getSafeBoolean(name: String): Boolean {
        return try {
            getBoolean(name)
        } catch (exception: JSONException) {
            false
        }
    }

    private fun JSONObject.getSafeInt(name: String): Int? {
        return try {
            getInt(name)
        } catch (exception: JSONException) {
            null
        }
    }

    private fun <T> JSONArray.iterable() = Iterable {
        object : Iterator<T> {
            var index = 0

            override fun next(): T {
                val element = get(index)
                index++
                return element as T
            }

            override fun hasNext(): Boolean {
                return index < length()
            }
        }
    }
}
